﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using UnityEngine;
using static Unity.XR.Qiyu.QiyuXRCorePlugin;

namespace Unity.XR.Qiyu.Space
{
    public class QiyuSpaceManager
    {
        #region RoomSetup
        private const string pluginName = "QiyuXRCore";

        [DllImport(pluginName, CallingConvention = CallingConvention.Cdecl)]
        public static extern void QVR_Space_InitRoom(IntPtr callback);
        [DllImport(pluginName, CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr QVR_Space_GetEntities(int queryMode, ref int count);

        [DllImport(pluginName, CallingConvention = CallingConvention.Cdecl)]
        public static extern Vector3f QVR_Relocator_ParseToUnityPoint(Vector3f pos);
        #endregion

        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        public delegate void SpaceCallBack(int code);

        [AOT.MonoPInvokeCallback(typeof(SpaceCallBack))]
        public static void OnSpaceCallBack(int code)
        {
            Debug.Log("OnSpaceCallBack:" + (SpaceCode)code);

            if (callbackInner != null)
            {
                callbackInner((SpaceCode)code);
                callbackInner = null;
            }
        }

        private static Action<SpaceCode> callbackInner;

        public static Vector3 Relocator_ParseToUnityPoint(Vector3 pos)
        {
            if (!QiyuPlatform.IsAndroid)
            {
                return pos;
            }

            Vector3f v3 = QVR_Relocator_ParseToUnityPoint(new Vector3f() { x = pos.x, y = pos.y, z = pos.z });
            return new Vector3() { x = v3.x, y = v3.y, z = v3.z };
        }


        public enum SpaceCode : int
        {
            Success = 0,
            RoomNotExist = 1,
            FindRoomFail = 2,
        }

        public enum QueryMode : int
        {
            QueryAllEntities = 0,
            Query2DEntities = 1,
            Query3DEntities = 2,
            QueryRoom = 3
        }

        public static void InitRoom(Action<SpaceCode> callback)
        {
            if (callbackInner != null)
                return;

            callbackInner = callback;
            SpaceCallBack callback_delegate = OnSpaceCallBack;
            QVR_Space_InitRoom(Marshal.GetFunctionPointerForDelegate(callback_delegate));
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct SpaceEntity
        {
            public long id;
            public int type;
            public Posef pose;
            public Rect3Df rect;
            public int pointCount;
            public IntPtr points;

            public override string ToString()
            {
                return $"id:{id},tag:{(SpaceEntityType)type},pose:{pose},rect:{rect},pointCount:{pointCount}";
            }
        }

        private static T[] GetNativeArray<T>(IntPtr array, int length)
        {
            T[] result = new T[length];

            int size = Marshal.SizeOf(typeof(T));
            if (IntPtr.Size == 4)
            {
                // 32-bit system
                for (int i = 0; i < result.Length; i++)
                {
                    result[i] = Marshal.PtrToStructure<T>(array);
                    array = new IntPtr(array.ToInt32() + size);
                }
            }
            else
            {
                //64-bit system
                for (int i = 0; i < result.Length; i++)
                {
                    result[i] = Marshal.PtrToStructure<T>(array);
                    array = new IntPtr(array.ToInt64() + size);
                }
            }
            return result;
        }

        public static List<SpaceAnchor> LoadSpaceModels()
        {
            List<SpaceAnchor> list = new List<SpaceAnchor>();
            if (!QiyuPlatform.IsAndroid) return list;

            LoadRoomLayout(list);
            LoadRoomParts_Quad(list);
            LoadRoomParts_Cube(list);
            return list;
        }

        private static SpaceRoomQuad ConvertQuadData(SpaceEntity data)
        {
            Debug.Log($"ConvertQuadData->{data}");
            SpaceRoomQuad quad = new SpaceRoomQuad();
            quad.id = data.id;
            quad.tag = (SpaceEntityType)data.type;
            quad.position = data.pose.Position.FromVector3f();
            quad.rotation = data.pose.Orientation.FromQuatf();
            quad.scale = data.rect.extent.FromExtent3Df();
            Vector3f[] points = GetNativeArray<Vector3f>(data.points, data.pointCount);
            for (int i = 0; i < points.Length; i++)
            {
                quad.boundary.Add(new Vector2(points[i].x, points[i].z));
                Debug.Log($"boundary idx:{i} point=>{ points[i]}");
            }

            return quad;
        }

        private static SpaceRoomCube ConvertCubeData(SpaceEntity data)
        {
            Debug.Log($"ConvertCubeData->{data}");
            SpaceRoomCube cube = new SpaceRoomCube();
            cube.id = data.id;
            cube.tag = (SpaceEntityType)data.type;
            cube.position = data.pose.Position.FromVector3f();
            cube.rotation = data.pose.Orientation.FromQuatf();
            cube.scale = data.rect.extent.FromExtent3Df();

            return cube;
        }

        private static void LoadRoomLayout(List<SpaceAnchor> list)
        {
            int count = 0;
            QVR_Space_GetEntities((int)QueryMode.QueryRoom, ref count);
            if (count > 0)
            {
                SpaceRoomLayout layout = new SpaceRoomLayout();
                Debug.Log("LoadRoomLayout count:" + count);
                IntPtr pBuff = QVR_Space_GetEntities((int)QueryMode.QueryRoom, ref count);
                SpaceEntity[] quads = GetNativeArray<SpaceEntity>(pBuff, count);
                for (int i = 0; i < quads.Length; i++)
                {

                    if (quads[i].type == (int)SpaceEntityType.Floor)
                    {
                        //floor
                        layout.floor = ConvertQuadData(quads[i]);
                    }
                    else if (quads[i].type == (int)SpaceEntityType.Ceiling)
                    {
                        //ceiling
                        layout.ceiling = ConvertQuadData(quads[i]);
                    }
                    else
                    {
                        layout.wallFaces.Add(ConvertQuadData(quads[i]));
                    }
                }

                list.Add(layout);
            }

        }

        public static void LoadRoomParts_Quad(List<SpaceAnchor> list)
        {
            int count = 0;
            QVR_Space_GetEntities((int)QueryMode.Query2DEntities, ref count);
            if (count > 0)
            {
                Debug.Log("LoadRoomParts_Quad count:" + count);
                IntPtr pBuff = QVR_Space_GetEntities((int)QueryMode.Query2DEntities, ref count);
                SpaceEntity[] quads = GetNativeArray<SpaceEntity>(pBuff, count);
                for (int i = 0; i < quads.Length; i++)
                {
                    list.Add(ConvertQuadData(quads[i]));
                }
            }
        }

        public static void LoadRoomParts_Cube(List<SpaceAnchor> list)
        {
            int count = 0;
            QVR_Space_GetEntities((int)QueryMode.Query3DEntities, ref count);
            if (count > 0)
            {
                Debug.Log("LoadRoomParts_Cube count:" + count);
                IntPtr pBuff = QVR_Space_GetEntities((int)QueryMode.Query3DEntities, ref count);
                SpaceEntity[] cubes = GetNativeArray<SpaceEntity>(pBuff, count);
                for (int i = 0; i < count; i++)
                {
                    list.Add(ConvertCubeData(cubes[i]));
                }
            }
        }
    }
}